#===============================================================================
# __init__.py
#===============================================================================

"""Initialization

This file contains the application factory for the acdumo app.
"""




# Imports ======================================================================

import os

from flask import Flask, render_template
from flask_login import login_required




# Functions ====================================================================

def create_app(test_config=None, configure_scheduler=True):
    """The application factory function

    This function creates and configures the Flask application object. For
    more on application factories, see the Flask documentation/tutorial:

    http://flask.pocoo.org/docs/1.0/tutorial/factory/

    http://flask.pocoo.org/docs/1.0/patterns/appfactories/

    Parameters
    ----------
    test_config : dict
        A dictionary containing configuration parameters for use during unit
        testing. If this parameter is `None`, the configuration will be loaded
        from `config.py` in the instance folder.

    Returns
    -------
    Flask
        A flask app
    """

    # create and configure the app
    app = Flask(__name__, instance_relative_config=True)
    
    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)
    
    # ensure the instance folder exists
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass
    
    from acdumo.models import db, migrate
    from acdumo.login import login
    from acdumo.email import mail
    from acdumo.errors import forbidden
    from acdumo.misaka import md
    for ext in db, login, mail, md:
        ext.init_app(app)
    migrate.init_app(app, db)
    login.login_view = 'auth.login'

    if configure_scheduler:
        from acdumo.apscheduler import scheduler
        scheduler.init_app(app)
        scheduler.start()

    for error, handler in ((403, forbidden),):
        app.register_error_handler(403, forbidden)
    
    from acdumo import auth, strategy, protected, settings
    for bp in auth.bp, strategy.bp, protected.bp, settings.bp:
        app.register_blueprint(bp)
    
    app.add_url_rule('/', endpoint='auth.login')
    
    return app



#===============================================================================
# acdumo.py
#===============================================================================

# Imports ======================================================================

import misaka as m
import os
import pandas as pd
import statistics
import seaborn as sns

from argparse import ArgumentParser
from datetime import datetime, timedelta
from operator import itemgetter
from yahoofinancials import YahooFinancials




# Constants ====================================================================

TICKERS = ['SPY', 'TLT', 'VSS', 'SCZ']
BONDS = 'TLT'
REPORT = """Date
====
{date}

Prices
======
![prices plot](prices.svg)

One Month Returns
=================
| Ticker | Return |
| ------ | ------ |
{returns}

Signals
=======
| Ticker | Signal |
| ------ | ------ |
{signals}

Strategy
========
{strategy}
"""




# Functions ====================================================================

def download_historical_price_data(date, *tickers, freq: str = 'monthly'):
    yahoo_financials = YahooFinancials(tickers)
    if freq == 'monthly':
        days = 217
    elif freq == 'weekly':
        days = 186

    def download_into_dfs(buffer=0):
        historical_price_data = yahoo_financials.get_historical_price_data(
            (date - timedelta(days=days)).strftime('%Y-%m-%d'),
            (date + timedelta(days=buffer)).strftime('%Y-%m-%d'),
            freq
        )
        return [
            pd.DataFrame(historical_price_data[ticker]['prices'][::-1])
            for ticker in tickers
        ]

    dfs = download_into_dfs()
    if freq == 'monthly':
        buffer = 0
        while date.strftime('%Y-%m-01') not in set(dfs[0].formatted_date):
            buffer += 1
            dfs = download_into_dfs(buffer=buffer)
    elif freq == 'weekly':
        dfs = [df.drop(0) for df in dfs]
    return {
        ticker: (
            df[
                [
                    d.endswith('-01') if freq == 'monthly' else True
                    for d in df.formatted_date
                ]
            ]
            .reset_index().dropna()
        )
        for ticker, df in zip(tickers, dfs)
    }


def plot_prices(historical_price_data: dict, file_name: str):
    hpd = pd.concat(
        df.assign(
            normalized_adjclose=df.adjclose / df.adjclose.iloc[-1],
            ticker=[ticker] * len(df.index)
        )[['formatted_date', 'normalized_adjclose', 'ticker']]
        for ticker, df in historical_price_data.items()
    )
    ax = sns.lineplot(
        x='formatted_date',
        y='normalized_adjclose',
        hue='ticker',
        data=hpd
    )
    ax.set_xlabel('')
    ax.set_ylabel(f'adjclose relative to {hpd.formatted_date.iloc[-1]}')
    ax.set_xticklabels(labels=hpd.formatted_date[::-1], rotation=30, ha='right')
    if len(ax.get_xticklabels()) > 8:
        for ind, label in enumerate(ax.get_xticklabels()):
            if ind % 4 == 0:
                label.set_visible(True)
            else:
                label.set_visible(False)
    fig = ax.get_figure()
    fig.tight_layout()
    fig.savefig(file_name, format='svg')
    fig.clf()


def compute_one_month_return(df, freq: str = 'monthly'):
    if freq == 'monthly':
        return df.adjclose[0] / df.adjclose[1] - 1
    elif freq == 'weekly':
        return (
            df.adjclose[0] / df.adjclose[4] - 1
        )


def compute_one_month_returns(
    historical_price_data: dict,
    freq: str = 'monthly'
):
    return {
        ticker: compute_one_month_return(df, freq=freq)
        for ticker, df in historical_price_data.items()
    }


def compute_signal(df, freq: str = 'monthly'):
    if freq == 'monthly':
        return sum(df.adjclose[0] / df.adjclose[x] - 1 for x in (1, 3, 6))
    elif freq == 'weekly':
        current_month_price = df.adjclose[0]
        return sum(
            current_month_price / df.adjclose[x] - 1
            for x in (4, 12, 24)
        )


def compute_signals(historical_price_data: dict, freq: str = 'monthly'):
    return {
        ticker: compute_signal(df, freq=freq)
        for ticker, df in historical_price_data.items()
    }


def decide_strategy(signals: dict, bonds: str = BONDS):
    signals_sans_bonds = dict((t, s) for t, s in signals.items() if t != bonds)
    if any(s > 0 for s in signals_sans_bonds.values()):
        choice = max(signals_sans_bonds.items(), key=itemgetter(1))[0]
    else:
        choice = bonds
    return f'Buy/Hold {choice}'


def generate_report(date, returns: dict, signals: dict, strategy: str):
    return REPORT.format(
        date=date.strftime('%Y-%m-%d'),
        returns='\n'.join(
            f'|    {t} | {s*100:.3}% |' for t, s in returns.items()
        ),
        signals='\n'.join(
            f'|    {t} | {s*100:.4}% |' for t, s in signals.items()
        ),
        strategy=strategy
    )


def parse_arguments():
    parser = ArgumentParser(description='Accelerated dual momentum')
    parser.add_argument(
        'report_dir',
        metavar='<path/to/report/dir/>',
        nargs='?',
        help='write a HTML report'
    )
    parser.add_argument(
        '--date',
        metavar='<yyyy-mm-dd>',
        default=datetime.today().strftime('%Y-%m-%d'),
        help=f'date of interest (default: today)'
    )
    parser.add_argument(
        '--tickers',
        metavar='<TIC>',
        nargs='+',
        default=TICKERS,
        help=f"tickers to use (default: {' '.join(TICKERS)})"
    )
    parser.add_argument(
        '--bonds',
        metavar='<TIC>',
        default=BONDS,
        help=f"ticker representing bonds (default: {BONDS})"
    )
    parser.add_argument(
        '--frequency',
        choices=('monthly', 'weekly'),
        default='monthly',
        help='frequency of data to fetch (default: monthly)'
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    date = datetime.strptime(args.date, '%Y-%m-%d')
    if date > datetime.today():
        raise RuntimeError(
            "I can't predict the future! Choose an earlier date."
        )
    historical_price_data = download_historical_price_data(
        date, *args.tickers, freq=args.frequency
    )
    returns = compute_one_month_returns(
        historical_price_data, freq=args.frequency
    )
    signals = compute_signals(historical_price_data, freq=args.frequency)
    strategy = decide_strategy(signals, bonds=args.bonds)
    report_text = generate_report(date, returns, signals, strategy)
    print(report_text, end='')
    if args.report_dir:
        if not os.path.isdir(args.report_dir):
            os.mkdir(args.report_dir)
        for ticker, df in historical_price_data.items():
            df.to_csv(
                os.path.join(args.report_dir, f'{ticker}.csv'),
                index=False
            )
        plot_prices(
            historical_price_data,
            os.path.join(args.report_dir, 'prices.svg')
        )
        with open(os.path.join(args.report_dir, 'acdumo.html'), 'w') as f:
            f.write(m.html(report_text, extensions=['tables']))




# Execute ======================================================================

if __name__ == '__main__':
    main()



#===============================================================================
# models.py
#===============================================================================

"""Databse models

Attributes
----------
db : SQLAlchemy
    The database object (see Flask-SQLAlchemy:
    http://flask-sqlalchemy.pocoo.org/2.3/ )
migrate : Migrate
    The Migrate object (see Flask-Migrate:
    https://flask-migrate.readthedocs.io/en/latest/ )
"""




# Imports ======================================================================

import jwt

from time import time
from datetime import datetime
from flask import g, current_app
from flask_login import UserMixin
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from werkzeug.security import generate_password_hash, check_password_hash




# Database initialization ======================================================

db = SQLAlchemy()
migrate = Migrate()




# Classes ======================================================================

# Models -----------------------------------------------------------------------



class User(UserMixin, db.Model):
    """A user

    Attributes
    ----------
    id : int
    username : str
    email : str
    password_hash : str
    email_confirmation_sent_on : datetime
    email_confirmed : bool
    email_confirmed_on : datetime
    """

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(64), index=True, unique=True)
    email = db.Column(db.String(120), index=True, unique=True)
    password_hash = db.Column(db.String(128))
    email_confirmation_sent_on = db.Column(db.DateTime, nullable=True)
    email_confirmed = db.Column(db.Boolean, default=False, nullable=True)
    email_confirmed_on = db.Column(db.DateTime, nullable=True)
    subscribed = db.Column(db.Boolean, default=False, nullable=True)

    def __repr__(self):
        """String representation of the user
        
        Returns
        -------
        str
            String representation of the user
        """

        return f'<User {self.username}>'
    
    def set_password(self, password):
        """Set the user's password

        Parameters
        ----------
        password : str
            The password to hash
        """

        self.password_hash = generate_password_hash(password)
    
    def check_password(self, password):
        """Check the provided password against the user's database record
        
        Parameters
        ----------
        password : str
            The provided password
        
        Returns
        -------
        bool
            The result of the hash check
        """

        return check_password_hash(self.password_hash, password)
    
    def get_confirm_email_token(self, expires_in=600):
        """Generate an email confirmation token

        Parameters
        ----------
        expires_in : int
            The lifetime of the token in seconds
        
        Returns
        -------
        bytes
            The token
        """

        return (
            jwt.encode(
                {'confirm_email': self.id, 'exp': time() + expires_in},
                current_app.config['SECRET_KEY'],
                algorithm='HS256'
            )
            .decode('utf-8')
        )
    
    @staticmethod
    def verify_confirm_email_token(token):
        """Verify an email confirmation token

        Parameters
        ----------
        token
            The token to be verified
        
        Returns
        -------
        User or None
            The user corresponding to the token if it is successfully verified,
            otherwise None.
        """
        
        try:
            id = jwt.decode(
                token,
                current_app.config['SECRET_KEY'],
                algorithms=['HS256']
            )['confirm_email']
        except:
            return
        return User.query.get(id)
    
    def get_reset_password_token(self, expires_in=600):
        """Generate a password reset token

        Parameters
        ----------
        expires_in : int
            The lifetime of the token in seconds
        
        Returns
        -------
        bytes
            The token
        """

        return (
            jwt.encode(
                {'reset_password': self.id, 'exp': time() + expires_in},
                current_app.config['SECRET_KEY'],
                algorithm='HS256'
            )
            .decode('utf-8')
        )

    @staticmethod
    def verify_reset_password_token(token):
        """Verify a password reset token

        Parameters
        ----------
        token
            The token to be verified
        
        Returns
        -------
        User or None
            The user corresponding to the token if it is successfully verified,
            otherwise None.
        """

        try:
            id = jwt.decode(
                token,
                current_app.config['SECRET_KEY'],
                algorithms=['HS256']
            )['reset_password']
        except:
            return
        return User.query.get(id)




# Functions ====================================================================

def get_db():
    """Retrieve the database from the application context
    
    Returns
    -------
    SQLAlchemy
        The database object
    """

    if 'db' not in g:
        g.db = db
    return g.db



#===============================================================================
# protected.py
#===============================================================================

"""Serve protected files

Attributes
----------
bp : Blueprint
    blueprint object, see the flask tutorial/documentation:

    http://flask.pocoo.org/docs/1.0/tutorial/views/

    http://flask.pocoo.org/docs/1.0/blueprints/
"""




# Imports ======================================================================

import os

from flask import Blueprint, send_from_directory, current_app
from flask_login import login_required




# Blueprint assignment =========================================================

bp = Blueprint('protected', __name__, url_prefix='/protected')




# Functions ====================================================================

@bp.route('/<path:filename>')
@login_required
def protected(filename):
    """Serve a protected file
    
    Parameters
    ----------
    filename
        Location of the file to serve on disk
    """

    return send_from_directory(
        os.path.join(current_app.instance_path, 'protected'),
        filename
    )



#===============================================================================
# settings.py
#===============================================================================

"""Settings blueprint

Attributes
----------
bp : Blueprint
    blueprint object, see the flask tutorial/documentation:

    http://flask.pocoo.org/docs/1.0/tutorial/views/

    http://flask.pocoo.org/docs/1.0/blueprints/
"""




# Imports ======================================================================

from flask import (
    Blueprint, render_template, current_app, flash, redirect, url_for
)
from flask_login import current_user, login_required
from werkzeug.exceptions import abort

from acdumo.models import get_db, User
from acdumo.forms import SubscribeForm, UnsubscribeForm




# Blueprint assignment =========================================================

bp = Blueprint('settings', __name__, url_prefix='/settings')




# Functions ====================================================================

@bp.route('/index', methods=('GET', 'POST'))
@login_required
def index():
    """Render the settings index"""
    
    user = User.query.get(current_user.id)
    if not user:
        flash('Strange, no account found.', 'error')
    if user.subscribed:
        form = UnsubscribeForm()
        if form.validate_on_submit():
            user.subscribed = False
            db = get_db()
            db.session.add(user)
            db.session.commit()
            flash('You have unsubscribed from acdumo email alerts')
            return redirect(url_for('settings.index'))
    else:
        form = SubscribeForm()
        if form.validate_on_submit():
            user.subscribed = True
            db = get_db()
            db.session.add(user)
            db.session.commit()
            flash('You are now subscribed to acdumo email alerts')
            return redirect(url_for('settings.index'))
    return render_template(
        'settings/index.html',
        form=form,
        not_subscribed=not user.subscribed
    )



#===============================================================================
# forms.py
#===============================================================================

"""Forms (subclasses of FlaskForm, see Flask-WTF:
http://flask-wtf.readthedocs.io/en/stable/ )
"""




# Imports  =====================================================================

from datetime import datetime

from flask import request, current_app
from flask_wtf import FlaskForm
from flask_wtf.file import FileField, FileAllowed
from wtforms import (
    StringField, PasswordField, BooleanField, SubmitField, TextAreaField
)
from wtforms.validators import (
    ValidationError, DataRequired, Email, EqualTo, Length
)
from acdumo.models import User




# Classes ======================================================================

# Forms ------------------------------------------------------------------------

class LoginForm(FlaskForm):
    """A form for user login credentials

    Attributes
    ----------
    username : StringField
    password : PasswordField
    remember_me : BooleanField
    submit : SubmitField
    """

    username = StringField('Username', validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired()])
    remember_me = BooleanField('Remember Me')
    submit = SubmitField('Log In')


class RegistrationForm(FlaskForm):
    """A form for user registration info

    Attributes
    ----------
    username : StringField
    email : StringField
    password : PasswordField
    password2 : PasswordField
    submit : SubmitField
    """

    username = StringField('Username', validators=[DataRequired()])
    email = StringField('Email', validators=[DataRequired(), Email()])
    password = PasswordField('Password', validators=[DataRequired()])
    password2 = PasswordField(
        'Repeat Password',
        validators=[DataRequired(), EqualTo('password')]
    )
    submit = SubmitField('Register')

    def validate_username(self, username):
        user = User.query.filter_by(username=username.data).first()
        if user is not None:
            raise ValidationError('Please use a different username.')
    
    def validate_email(self, email):
        user = User.query.filter_by(email=email.data).first()
        if user is not None:
            raise ValidationError('Please use a different email address.')


class ResetPasswordRequestForm(FlaskForm):
    """A form for a password reset request

    Attributes
    ----------
    username : StringField
    email : StringField
    submit : SubmitField
    """

    username = StringField('Username', validators=[DataRequired()])
    email = StringField('Email', validators=[DataRequired(), Email()])
    submit = SubmitField('Request Password Reset')


class ResetPasswordForm(FlaskForm):
    """A form for providing a new password

    Attributes
    ----------
    password : PasswordField
    password2 : PasswordField
    submit : SubmitField
    """

    password = PasswordField('Password', validators=[DataRequired()])
    password2 = PasswordField(
        'Repeat Password',
        validators=[DataRequired(), EqualTo('password')]
    )
    submit = SubmitField('Request Password Reset')


class StrategyForm(FlaskForm):
    """A form for strategy queries

    Attributes
    ----------
    date : StringField
    tickers : StringField
    submit : SubmitField
    """

    date = StringField('yyyy-mm-dd', default=datetime.today().strftime('%Y-%m-%d'))
    tickers = StringField('TK1 TK2 TK3...', default='SPY TLT VSS SCZ')
    submit = SubmitField('Submit')


class SubscribeForm(FlaskForm):
    """A form for email subscription

    Attributes
    ----------
    tickers : StringField
    submit : SubmitField
    """
    
    tickers = StringField('TK1 TK2 TK3...', default='SPY TLT VSS SCZ')
    submit = SubmitField('Subscribe')


class UnsubscribeForm(FlaskForm):
    """A form for email subscription

    Attributes
    ----------
    submit : SubmitField
    """

    submit = SubmitField('Unsubscribe')



#!/user/bin/env python3
#===============================================================================
# config
#===============================================================================

"""Create a config file"""




# Imports ======================================================================

import argparse
import os
import os.path




# Constants ====================================================================

DEVELOPMENT_CONFIG_DATA = f'''
import os
import os.path
basedir = os.path.abspath(os.path.dirname(__file__))
SECRET_KEY = os.environ.get('SECRET_KEY') or {os.urandom(16)}
SQLALCHEMY_DATABASE_URI = (
    os.environ.get('DATABASE_URL')
    or 'sqlite:///' + os.path.join(basedir, 'app.db')
)
SQLALCHEMY_TRACK_MODIFICATIONS = False
PROTECTED_DIR = os.path.join(basedir, 'protected')
MAIL_SERVER = os.environ.get('MAIL_SERVER') or 'smtp.gmail.com'
MAIL_PORT = int(os.environ.get('MAIL_PORT') or 587)
MAIL_USE_TLS = True
MAIL_USE_SSL = False
MAIL_USERNAME = ''
MAIL_PASSWORD = os.environ.get('MAIL_PASSWORD')
ADMINS = ['anthony.aylward@protonmail.com']
APPROVED_EMAILS = ['anthony.aylward@protonmail.com']
'''

PRODUCTION_CONFIG_DATA = f'''
import os
import os.path
basedir = os.path.abspath(os.path.dirname(__file__))
SECRET_KEY = os.environ.get('SECRET_KEY') or {os.urandom(16)}
SQLALCHEMY_DATABASE_URI = (
    os.environ.get('DATABASE_URL')
    or 'sqlite:///' + os.path.join(basedir, 'app.db')
)
SQLALCHEMY_TRACK_MODIFICATIONS = False
PROTECTED_DIR = os.path.join(basedir, 'protected')
MAIL_SERVER = os.environ.get('MAIL_SERVER') or 'mail.smtp2go.com'
MAIL_PORT = int(os.environ.get('MAIL_PORT') or 465)
MAIL_USE_TLS = False
MAIL_USE_SSL = True
MAIL_USERNAME = 'accelerated.dual.momentum@acdumo.info'
MAIL_PASSWORD = os.environ.get('MAIL_PASSWORD')
ADMINS = ['accelerated.dual.momentum@acdumo.info']
APPROVED_EMAILS = [
    'anthony.aylward@protonmail.com',
    'green.danying@gmail.com',
    'craylward@gmail.com',
    'elijahwaichunkun@gmail.com',
    'work@lbry.com'
]
'''




# Functions ====================================================================

def parse_arguments():
    parser = argparse.ArgumentParser(
        description='write configuration file'
    )
    parser.add_argument(
        'instance',
        metavar='<path/to/instance-folder/>',
        help='path to instance folder'
    )
    parser.add_argument(
        '--production',
        action='store_true',
        help='write a production config file'
    )
    return parser.parse_args()

def main():
    args = parse_arguments()
    with open(os.path.join(args.instance, 'config.py'), 'w') as f:
        f.write(
            PRODUCTION_CONFIG_DATA if args.production
            else DEVELOPMENT_CONFIG_DATA
        )




# Execute ======================================================================

if __name__ == '__main__':
    main()



#===============================================================================
# auth.py
#===============================================================================

"""Authentication blueprint

Attributes
----------
bp : Blueprint
    blueprint object, see the flask tutorial/documentation:

    http://flask.pocoo.org/docs/1.0/tutorial/views/

    http://flask.pocoo.org/docs/1.0/blueprints/
"""




# Imports ======================================================================

from datetime import datetime
from flask import (
    Blueprint, flash, g, redirect, render_template, request, session, url_for,
    current_app
)
from flask_login import current_user, login_user, logout_user
from werkzeug.security import check_password_hash, generate_password_hash
from werkzeug.urls import url_parse
from acdumo.forms import (
    LoginForm, RegistrationForm, ResetPasswordRequestForm, ResetPasswordForm
)
from acdumo.models import get_db, User
from acdumo.email import (
    send_confirmation_email, send_password_reset_email
)




# Blueprint assignment =========================================================

bp = Blueprint('auth', __name__, url_prefix='/auth')




# Functions ====================================================================

@bp.route('/register', methods=('GET', 'POST'))
def register():
    """Register a new user
    
    If the current user is already logged in, they will be redirected to the
    index page.

    Otherwise, the registration page will be rendered. It includes a
    RegistrationForm (see `forms.py`).
    
    If the supplied email is on the approved email list (and does not already
    have an account, see models.User), a new user will be created from the form
    data and added to the database. There it will await confirmation (see
    `confirm_email`)
    """

    if current_user.is_authenticated:
        return redirect(url_for('strategy.index'))
    form = RegistrationForm()
    if form.validate_on_submit():
        if form.email.data not in current_app.config['APPROVED_EMAILS']:
            flash('Sorry, that email is not on the approved list.')
            return redirect(url_for('auth.login'))
        user = User(username=form.username.data, email=form.email.data)
        user.set_password(form.password.data)
        db = get_db()
        db.create_all()
        db.session.add(user)
        db.session.commit()
        send_confirmation_email(user)
        flash('Please check your email to confirm your email address.')
        return redirect(url_for('auth.login'))
    return render_template(
        'auth/register.html',
        title='Register',
        form=form
    )


@bp.route('/login', methods=('GET', 'POST'))
def login():
    """Log in to the site

    If the current user is already logged in, they will be redirected to the
    index page.

    Otherwise, the login page will be rendered. It includes a LoginForm (see
    `forms.py`). Supplying valid credentials will allow the user to log in.
    """

    if current_user.is_authenticated:
        return redirect(url_for('strategy.index'))
    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(username=form.username.data).first()
        if user is None or any(
            (
                not user.check_password(form.password.data),
                not user.email_confirmed
            )
        ):
            flash('Invalid username or password', 'error')
            return redirect(url_for('auth.login'))
        login_user(user, remember=form.remember_me.data)
        next_page = request.args.get('next')
        if not next_page or url_parse(next_page).netloc != '':
            next_page = url_for('strategy.index')
        return redirect(next_page)
    return render_template('auth/login.html', title='Log In', form=form)


@bp.route('/logout')
def logout():
    """Log out the current user"""

    logout_user()
    return redirect(url_for('auth.login'))


@bp.route('/confirm/<token>')
def confirm_email(token):
    """Email confirmation page

    If the current user is already logged in, they will be redirected to the
    index page.

    This function renders the page liked to by the registration confirmation
    email. It includes a message about the success or failure of the
    confirmation.

    Parameters
    ----------
    token
        The JSON web token
    """

    if current_user.is_authenticated:
        return redirect(url_for('strategy.index'))
    user = User.verify_confirm_email_token(token)
    if not user:
        flash('Strange, no account found.', 'error')
    if user.email_confirmed:
        flash('Account already confirmed. Please login.', 'info')
    else:
        user.email_confirmed = True
        user.email_confirmed_on = datetime.utcnow()
        db = get_db()
        db.session.add(user)
        db.session.commit()
        flash('Thank you for confirming your email address!')
    return redirect(url_for('auth.login'))


@bp.route('/reset_password_request', methods=('GET', 'POST'))
def reset_password_request():
    """Request a password reset

    If the current user is already logged in, they will be redirected to the
    index page.

    Otherwise, the reset password page will be rendered. It includes a
    ResetPasswordRequestForm (see `forms.py`). Submitting a valid
    username-email pair will cause a password reset email to be sent.
    """

    if current_user.is_authenticated:
        return redirect(url_for('strategy.index'))
    form = ResetPasswordRequestForm()
    if form.validate_on_submit():
        user = User.query.filter_by(username=form.username.data).first()
        if user and user.email == form.email.data:
            send_password_reset_email(user)
            flash(
                'Check your email for the instructions to reset your password'
            )
            return redirect(url_for('auth.login'))
        else:
            flash('Invalid username/email pair')
            return redirect(url_for('auth.reset_password_request'))
    return render_template(
        'auth/reset_password_request.html',
        title='Reset Password',
        form=form
    )


@bp.route('/reset_password/<token>', methods=('GET', 'POST'))
def reset_password(token):
    """Reset a user's password
    
    If the current user is already logged in, they will be redirected to the
    index page.

    This function renders the page linked to by the password reset email. The
    link includes a JSON web token as a variable component of the URL. If the
    token cannot be verified, the user is redirected to the login page.

    If the token is verified, the user's password will be reset according to
    the data entered into the included ResetPasswordForm (see `forms.py`).

    Parameters
    ----------
    token
        the JSON web token
    """

    if current_user.is_authenticated:
        return redirect(url_for('strategy.index'))
    user = User.verify_reset_password_token(token)
    if not user:
        return redirect(url_for('auth.login'))
    form = ResetPasswordForm()
    if form.validate_on_submit():
        user.set_password(form.password.data)
        db = get_db()
        db.session.commit()
        flash('Your password has been reset.')    
        return redirect(url_for('auth.login'))
    return render_template('auth/reset_password.html', form=form)



#===============================================================================
# email.py
#===============================================================================

"""Instantiation of the Mail object, along with helper functions for email
handling

Attributes
----------
mail : Mail
    A Mail object (see Flask-Mail: https://pythonhosted.org/Flask-Mail/ )
"""




# Imports ======================================================================

from flask import render_template, current_app
from flask_mail import Mail, Message
from threading import Thread

from acdumo import create_app





# Initialization ===============================================================

mail = Mail()




# Functions ====================================================================

def send_async_email(app, msg):
    """Send an email asynchronously

    An application context is pushed so that the mail operation can use its own
    thread. See the Flask documentation for more on application contexts:

    http://flask.pocoo.org/docs/1.0/appcontext/

    Parameters
    ----------
    app : Flask
        The application from which to push a context
    msg : Message
        The message to be sent
    """

    with app.app_context():
        mail.send(msg)

def send_email(subject, sender, recipients, text_body, html_body):
    """Send an email

    A Message object is instantiated using the provided subject, sender and
    recippients, then filled using the provided text/html data. A new app is
    created to support sending mail asynchrounously, and finally
    send_async_email is called with its own thread.

    Parameters
    ----------
    subject
        The subject of the email
    sender
        The sender of the email
    recipients
        The recipients of the email
    text_body
        The body of the email (text version)
    html_body
        The body of the email (html version)
    """

    msg = Message(subject, sender=sender, recipients=recipients)
    msg.body = text_body
    msg.html = html_body
    app = create_app(configure_scheduler=False)
    Thread(target=send_async_email, args=(app, msg)).start()


def send_password_reset_email(user):
    """Send a password reset email

    A JSON web token is created, then included in an email to the account for
    which a password reset has been requested.

    Parameters
    ----------
    user : User
        User whose password is to be reset
    """

    token = user.get_reset_password_token()
    send_email(
        '[acdumo] Reset Your Password',
        sender=current_app.config['ADMINS'][0],
        recipients=[user.email],
        text_body=render_template(
            'email/reset_password.txt',
            user=user,
            token=token
        ),
        html_body=render_template(
            'email/reset_password.html',
            user=user,
            token=token
        )
    )


def send_confirmation_email(user):
    """Send a confirmation email

    A JSON web token is created, then included in an email to the account for
    which the email is to be confirmed

    Parameters
    ----------
    user : User
        The user whose email is to be confirmed
    """

    token = user.get_confirm_email_token()
    send_email(
        '[acdumo] Confirm Your Email Address',
        sender=current_app.config['ADMINS'][0],
        recipients=[user.email],
        text_body=render_template(
            'email/email_confirmation.txt',
            user=user,
            token=token
        ),
        html_body=render_template(
            'email/email_confirmation.html',
            user=user,
            token=token
        )
    )


def send_notification_email(user, date, report, alert):
    """Send a password reset email

    An email with a signal update is sent

    Parameters
    ----------
    user : User
        User to whom email will be sent
    date
        date for the report
    report
        report to send
    """

    send_email(
        '[acdumo] signal notification',
        sender=current_app.config['ADMINS'][0],
        recipients=[user.email],
        text_body=render_template(
            'email/notification.txt',
            date=date,
            report=report,
            alert=alert
        ),
        html_body=render_template(
            'email/notification.html',
            date=date,
            report=report,
            alert=alert
        )
    )



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='acdumo',
    version='1.0.21',
    author='Anthony Aylward, Serena Wu, Elijah Kun',
    author_email='aaylward@eng.ucsd.edu',
    description='Accelerated dual momentum',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/anthony-aylward/acdumo',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires = [
        'flask',
        'flask-apscheduler',
        'flask-login',
        'flask-mail',
        'flask-migrate',
        'flask-misaka',
        'flask-wtf',
        'pyjwt',
        'misaka',
        'pandas',
        'seaborn',
        'yahoofinancials',
    ],
    entry_points = {
        'console_scripts': [
            'acdumo=acdumo.acdumo:main',
            'acdumo-install-certifi=acdumo.install_certifi:main'
        ]
    }
)



from flask import g
from datetime import datetime, timedelta
from flask_apscheduler import APScheduler
from acdumo import create_app
from acdumo.email import send_notification_email
from acdumo.models import get_db, User
from acdumo.acdumo import (
    TICKERS, BONDS, download_historical_price_data,
    compute_signals, decide_strategy, plot_prices
)

scheduler = APScheduler()

REPORT = """Date
====
{date}

Signals
=======
| Ticker | Signal |
| ------ | ------ |
{signals}

Strategy
========
{strategy}
"""


@scheduler.task(
    'cron',
    id='send_notification_email',
    day_of_week='mon,tue,wed,thu,fri',
    hour='7',
    misfire_grace_time=900
)
def notification_email():
    date = datetime.today()
    hpd_yesterday = download_historical_price_data(date - timedelta(days=1), *TICKERS, freq='weekly')
    signals_yesterday = compute_signals(hpd_yesterday, freq='weekly')
    strategy_yesterday = decide_strategy(signals_yesterday, bonds=BONDS)
    hpd = download_historical_price_data(date, *TICKERS, freq='weekly')
    signals = compute_signals(hpd, freq='weekly')
    strategy = decide_strategy(signals, bonds=BONDS)
    formatted_date = datetime.today().strftime('%Y-%m-%d')
    signals_sans_bonds = dict((t, s) for t, s in signals.items() if t != BONDS)
    alert = None
    if strategy != strategy_yesterday:
        alert = 'A strategy change has occurred recently.'
    elif any(abs(signals[BONDS] - signals[s]) < 0.1 for s in signals_sans_bonds.keys()):
        alert = (
            'A stocks signal is within 10% of the bonds signal, so a '
            'strategy change may occur soon.'
        )
    if alert is not None:
        app = create_app(configure_scheduler=False)
        with app.app_context():
            db = get_db()
            for user in User.query.all():
                if user.subscribed:
                    send_notification_email(
                        user,
                        formatted_date,
                        REPORT.format(
                            date=formatted_date,
                            signals='\n'.join(
                                f'|    {t} | {s*100:.4}% |' for t, s in signals.items()
                            ),
                            strategy=strategy
                        ),
                        alert=alert
                    )



#===============================================================================
# login.py
#===============================================================================

"""Instantiation of the LoginManager (see Flask-Login: 
http://flask-login.readthedocs.io/en/latest/#configuring-your-application )

Attributes
----------
login : LoginManager
    The login manager
"""




# Imports ======================================================================

from flask_login import LoginManager
from acdumo.models import User




# Initialization ===============================================================

login = LoginManager()




# Functions ====================================================================

@login.user_loader
def load_user(id):
    """The load_user function required by the login manager
    
    This tells the login manager how to load a user. In this case, the user
    is retrieved from the database by ID no.

    Parameters
    ----------
    id
        The ID no. of the user to be loaded
    
    Returns
    -------
    User | WhisperUser
        The user to be loaded
    """

    return User.query.get(int(id))



#===============================================================================
# strategy.py
#===============================================================================

"""Strategy blueprint

Attributes
----------
bp : Blueprint
    blueprint object, see the flask tutorial/documentation:

    http://flask.pocoo.org/docs/1.0/tutorial/views/

    http://flask.pocoo.org/docs/1.0/blueprints/
"""




# Imports ======================================================================

import os.path

from datetime import datetime
from flask import (
    Blueprint, render_template, current_app, url_for, redirect, request
)
from flask_login import login_required

from acdumo.forms import StrategyForm
from acdumo.acdumo import (
    TICKERS, BONDS, download_historical_price_data, compute_one_month_returns,
    compute_signals, decide_strategy, plot_prices
)



# Blueprint assignment =========================================================

bp = Blueprint('strategy', __name__)




# Constants ====================================================================

REPORT = """Date
----
{date}

Strategy
--------
{strategy}

Signals
-------
| Ticker | Signal |
| ------ | ------ |
{signals}

One Month Returns
-----------------
| Ticker | Return |
| ------ | ------ |
{returns}

Prices
------
![prices plot]({prices_svg})

CSV Data
--------
{csv_links}
"""




# Functions ====================================================================


@bp.route('/', methods=('GET', 'POST'))
@login_required
def index():
    f"""Render the strategy index"""

    form = StrategyForm()
    if form.validate_on_submit():
        return redirect(
            url_for(
                'strategy.index',
                date=form.date.data,
                tickers=form.tickers.data
            )
        )

    date = datetime.strptime(
        request.args.get('date', datetime.today().strftime('%Y-%m-%d')),
        '%Y-%m-%d'
    )
    tickers = request.args.get('tickers', ' '.join(TICKERS)).split()

    hpd = download_historical_price_data(date, *tickers, freq='weekly')
    plot_prices(
        hpd,
        os.path.join(current_app.config['PROTECTED_DIR'], f"prices-{'-'.join(tickers)}-{date.strftime('%Y-%m-%d')}.svg")
    )
    returns = compute_one_month_returns(hpd, freq='weekly')
    signals = compute_signals(hpd, freq='weekly')
    strategy = decide_strategy(signals, bonds=BONDS)
    formatted_date = date.strftime('%Y-%m-%d')
    for ticker, df in hpd.items():
        df.to_csv(
            os.path.join(current_app.config['PROTECTED_DIR'], f"{ticker}-{formatted_date}.csv"),
            index=False
        )
    return render_template(
        'strategy/index.html',
        report=REPORT.format(
            date=formatted_date,
            strategy=strategy,
            signals='\n'.join(
                f'|    {t} | {s*100:.4}% |' for t, s in signals.items()
            ),
            returns='\n'.join(
                f'|    {t} | {s*100:.3}% |' for t, s in returns.items()
            ),
            prices_svg=url_for(
                'protected.protected',
                filename=f"prices-{'-'.join(tickers)}-{formatted_date}.svg"
            ),
            csv_links='\n'.join(
                f"<a href=\"{url_for('protected.protected',filename=f'{ticker}-{formatted_date}.csv')}\" class='btn btn-outline-primary'>{ticker}</a>"
                for ticker in tickers
            )
        ),
        form=form,
        date=formatted_date,
        tickers=' '.join(tickers)
    )
